#include <rk/rk_linux.h>
#include <rk/rk_error.h>
#include <rk/rk.h>
#include <linux/time.h>
#include <asm/processor.h>

static	int		cpu_reserve_destroy(rk_reserve_t);
static	void		cpu_reserve_start_account(rk_reserve_t);
static	void		cpu_reserve_stop_account(rk_reserve_t, cpu_tick_t);
static	void		cpu_reserve_replenish(rk_reserve_t, cpu_tick_t);
static	void		cpu_reserve_enforce(rk_reserve_t);
static	void		cpu_reserve_sleep_on(rk_reserve_t);
static	void		cpu_reserve_attach(rk_reserve_t, struct rs_proc_list*);
static	void		cpu_reserve_detach(rk_reserve_t, struct rs_proc_list*);
static	int		cpu_reserve_read_proc(rk_reserve_t, char *);

void			cpu_reserve_sched_realtime(struct rs_proc_list*);
void			cpu_reserve_sched_timeshare(struct rs_proc_list*);
void			cpu_reserve_sched_timeshare_self(void);

#ifdef	UPERF_RK
#define	MAX_COUNT	1000
static	unsigned long	*result_start_account;
static	unsigned long	counter_start_account;
static	unsigned long	*result_start_account_timer;
static	unsigned long	counter_start_account_timer;
static	unsigned long	*result_stop_account;
static	unsigned long	counter_stop_account;
static	unsigned long	*result_replenish;
static	unsigned long	counter_replenish;
static	unsigned long	*result_enforce;
static	unsigned long	counter_enforce;

int	uperf_rk_isa_timer_programmed;

static void
cpu_reserve_perf_init(void)
{
	result_start_account = malloc(sizeof(unsigned long) * MAX_COUNT);
	result_start_account_timer = malloc(sizeof(unsigned long) * MAX_COUNT);
	result_stop_account = malloc(sizeof(unsigned long) * MAX_COUNT);
	result_replenish = malloc(sizeof(unsigned long) * MAX_COUNT);
	result_enforce = malloc(sizeof(unsigned long) * MAX_COUNT);

	bzero(result_start_account, sizeof(unsigned long) * MAX_COUNT);
	bzero(result_start_account_timer, sizeof(unsigned long) * MAX_COUNT);
	bzero(result_stop_account, sizeof(unsigned long) * MAX_COUNT);
	bzero(result_replenish, sizeof(unsigned long) * MAX_COUNT);
	bzero(result_enforce, sizeof(unsigned long) * MAX_COUNT);
}

static unsigned long
perf_average(unsigned long *result, unsigned long count)
{
	unsigned long	i;
	cpu_tick_data_t	total;

	if (count == 0)
		return 0;
	for (i = 0, total = 0; i < count; i++) {
		total += result[i];
	}
	return (unsigned long)(total/count);
}

#define	PROLOGUE_PERF			\
	cpu_tick_data_t	perf_start, perf_end; \
	rk_rdtsc(&perf_start);

#define	EPILOGUE_PERF(what)		\
	rk_rdtsc(&perf_end);		\
	result_##what[(counter_##what)++] = (unsigned long)(perf_end - perf_start);	\
	if ((counter_##what) >= MAX_COUNT) (counter_##what) = 0;

static void
cpu_reserve_start_account_perf(rk_reserve_t rsv)
{
	PROLOGUE_PERF;
	uperf_rk_isa_timer_programmed = 0;
	cpu_reserve_start_account(rsv);
	if (uperf_rk_isa_timer_programmed) {
		EPILOGUE_PERF(start_account_timer);
	} else {
		EPILOGUE_PERF(start_account);
	}
}

static void
cpu_reserve_stop_account_perf(rk_reserve_t rsv, cpu_tick_t now)
{
	PROLOGUE_PERF
	cpu_reserve_stop_account(rsv, now);
	EPILOGUE_PERF(stop_account)
}

static void
cpu_reserve_replenish_perf(rk_reserve_t rsv, cpu_tick_t period)
{
	PROLOGUE_PERF
	cpu_reserve_replenish(rsv, period);
	EPILOGUE_PERF(replenish)
}

static void
cpu_reserve_enforce_perf(rk_reserve_t rsv)
{
	PROLOGUE_PERF
	cpu_reserve_enforce(rsv);
	EPILOGUE_PERF(enforce)
}

struct rk_reserve_ops	cpu_reserve_ops = {
	cpu_reserve_destroy,
	cpu_reserve_start_account_perf,
	cpu_reserve_stop_account_perf,
	cpu_reserve_replenish_perf,
	cpu_reserve_enforce_perf,
	cpu_reserve_attach,
	cpu_reserve_detach,
#ifdef	linux
	cpu_reserve_sleep_on,
	cpu_reserve_read_proc,
#endif
};
#else	/* UPERF_RK*/
struct rk_reserve_ops	cpu_reserve_ops = {
	cpu_reserve_destroy,
	cpu_reserve_start_account,
	cpu_reserve_stop_account,
	cpu_reserve_replenish,
	cpu_reserve_enforce,
	cpu_reserve_attach,
	cpu_reserve_detach,
#ifdef	linux
	cpu_reserve_sleep_on,
	cpu_reserve_read_proc,
#endif
};
#endif	/* UPERF_RK*/

/*
 * Internal data structure to keep track of cpu capacity usage
 *	99.99% is represented as 9999
 */
typedef	unsigned long		cpu_capacity_t;
typedef	unsigned long long	cpu_capacity_quad_t;

#define	CAPACITY_INT(x)		((x)/100)
#define	CAPACITY_FRAC(x)	((x)-(x/100)*100)
#define	INT2CAPACITY(x)		((x)*10000)
#define	CAPACITY2INT(x)		((x)/10000)
#define	PERCENT2CAPACITY(x)	((x)*100)
#define	CAPACITY_OF(c,t)	(INT2CAPACITY(c)/(t))

/*
 *	66% ... is the maximum available capacity for reserves.
 */
#define	CPU_CAPACITY_MAX	PERCENT2CAPACITY(66)

/*
 *
 */
struct cpu_reserve {
	/*current period */
	cpu_tick_data_t	cpu_period_used_ticks;		/* CPU ticks already used 
							   in the current period */
	cpu_tick_data_t	cpu_period_start_ticks;		/* remember start ticks for 
							   the current invocation */
	cpu_tick_data_t	cpu_period_available_ticks;	/* CPU ticks that can be used 
							   for this period */
	struct wait_queue 	*depleted_wait;		/* for enforcement */
	/* execution statistics */
	cpu_tick_data_t		cpu_period_prev_used_ticks;
	cpu_capacity_t		cpu_max_utilization;
	cpu_capacity_t		cpu_min_utilization;
	struct {
		unsigned long		total_count;	/* total count of replenishment */
		cpu_capacity_quad_t	total_utils;	/* summation of utilizations */
	}			cpu_average;
	/* static data */
	struct list_head	cpu_link;		/* link all cpu reserves */
	struct timespec		cpu_time;
	struct timespec		cpu_period;
	cpu_tick_data_t		cpu_time_ticks;		/* ticks for cpu_time */
	cpu_tick_data_t		cpu_period_ticks;	/* ticks for cpu_period */
	cpu_capacity_t		cpu_capacity;		/* requested capacity */
#if	defined(EVAL_RK)
	unsigned long		cpu_eval_data_count;
	struct cpu_eval_data	*cpu_eval_datap;
#endif
};
typedef	struct cpu_reserve	*cpu_reserve_t;

#if	defined(EVAL_RK)
#define	MAX_COUNT	3000
extern	unsigned long volatile jiffies;
struct	cpu_eval_data {
	unsigned long		t_jiffies;		/* jiffies when recorded */
	unsigned long		c_used;			/* ticks used in period */
};
#endif
/*
 *
 */
struct list_head	cpu_reserve_head;
cpu_capacity_t		cpu_reserve_current_capacity;	/* for admission control */

/*
 *
 */
void
cpu_reserve_init(void)
{
	INIT_LIST_HEAD(&cpu_reserve_head);
	cpu_reserve_current_capacity = 0;
#ifdef	UPERF_RK
	cpu_reserve_perf_init();
#endif	/*UPERF_RK*/
}

rk_reserve_t
cpu_reserve_create(rk_resource_set_t rs,
		   struct timespec *c, struct timespec *t,
		   rk_reserve_param_t p)
{
	rk_reserve_t	rsv;
	cpu_reserve_t	cpu;
	cpu_capacity_t	capacity;
	cpu_capacity_t	new_capacity;
	cpu_capacity_quad_t	qc, qt;

#ifdef	DEBUG_RK
	printk("cpu_reserve_create: rs(0x%x) c(%d,%d) t(%d,%d)\n",
	       (int)rs,
	       (int)c->tv_sec, (int)c->tv_nsec, (int)t->tv_sec, (int)t->tv_nsec);
#endif

	/* admission control */
	qc = c->tv_sec; qc *= NANOSEC; qc += c->tv_nsec;
	qt = t->tv_sec; qt *= NANOSEC; qt += t->tv_nsec;

	capacity = CAPACITY_OF(qc, qt);
	new_capacity = capacity + cpu_reserve_current_capacity;

#ifdef	DEBUG_RK
	printk("cpu_reserve_create: capacity(%lu.%02lu) new total(%lu.%02lu)\n",
	       CAPACITY_INT(capacity), CAPACITY_FRAC(capacity),
	       CAPACITY_INT(new_capacity), CAPACITY_FRAC(new_capacity));
#endif

	if (new_capacity > CPU_CAPACITY_MAX) {
		/* admission control failed */
		return NULL_RESERVE;	/* error */
	}

	cpu_reserve_current_capacity = new_capacity;

	/* create cpu reserve object */
	cpu = malloc(sizeof(struct cpu_reserve));
	bzero(cpu, sizeof(struct cpu_reserve));
	cpu->cpu_time = *c;
	cpu->cpu_period = *t;
	cpu->cpu_capacity = capacity;
	list_add(&cpu->cpu_link, &cpu_reserve_head);

	/* calculate cpu ticks per capacity */
	nanosec2tick(&qt, &cpu->cpu_period_ticks);
	nanosec2tick(&qc, &cpu->cpu_time_ticks);

	cpu->cpu_period_available_ticks = cpu->cpu_time_ticks;

	/* init waitqueue */
	init_waitqueue(&cpu->depleted_wait);

#ifdef	EVAL_RK
	{
		int size = (rk_cpu_ticks_per_second/cpu->cpu_period_ticks)*180;	/* 3min */
		if (size > MAX_COUNT) {
			size = MAX_COUNT;
		}
		cpu->cpu_eval_data_count = 0;
		cpu->cpu_eval_datap = malloc(sizeof(struct cpu_eval_data)*size);
		bzero(cpu->cpu_eval_datap, sizeof(struct cpu_eval_data)*size);
	}
#endif

	/* create generic reserve object */
	rsv = rk_reserve_create(rs, RSV_CPU);
	rsv->rsv_state = RSV_IS_NULL;
	rsv->rsv_rsv = cpu;
	rsv->rsv_ops = &cpu_reserve_ops;
	rsv->rsv_reserve_param = *p;

	/* create a timer for it */
	rk_replenish_timer_create(rsv, &cpu->cpu_period_ticks);

	/* attach reserve to resource set*/
	rk_resource_set_attach_reserve(rs, rsv);

	return rsv;		/* success */
}

int
cpu_reserve_ctl(rk_resource_set_t rs, rk_reserve_t rsv,
		struct timespec *c, struct timespec *t)
{
	cpu_reserve_t	cpu;
	cpu_capacity_t	capacity;
	cpu_capacity_t	new_capacity;
	cpu_capacity_quad_t	qc, qt;

	if (rsv->rsv_rs != rs || rs->rs_cpu != rsv) {
		printk("cpu_reserve_ctl: rs(0x%x) rsv(0x%x) do not match.\n",
		       (int)rs, (int)rsv);
		return RK_ERROR;	/* error */
	}

	cpu = rsv->rsv_rsv;

#ifdef	DEBUG_RK
	printk("cpu_reserve_ctl: rsv(0x%x) OLD: c(%ld,%ld) t(%ld,%ld) NEW: c(%ld,%ld) t(%ld,%ld)\n",
	       (int)rsv,
	       cpu->cpu_time.tv_sec, cpu->cpu_time.tv_nsec,
	       cpu->cpu_period.tv_sec, cpu->cpu_period.tv_nsec,
	       c->tv_sec, c->tv_nsec, t->tv_sec, t->tv_nsec);
#endif

	/* calculate a new capacity */
	qc = c->tv_sec; qc *= NANOSEC; qc += c->tv_nsec;
	qt = t->tv_sec; qt *= NANOSEC; qt += t->tv_nsec;
	capacity = CAPACITY_OF(qc, qt);
	new_capacity = capacity + cpu_reserve_current_capacity - cpu->cpu_capacity;

#ifdef	DEBUG_RK
	printk("cpu_reserve_ctl: capacity(%lu.%02lu) new total(%lu.%02lu)\n",
	       CAPACITY_INT(capacity), CAPACITY_FRAC(capacity),
	       CAPACITY_INT(new_capacity), CAPACITY_FRAC(new_capacity));
#endif

	/* if it is less than the current capacity,  use the new parameters. */
	/* if not, do an admission control test */
	if (capacity > cpu->cpu_capacity) {
		if (new_capacity > CPU_CAPACITY_MAX) {
			/* admission control failed */
			return RK_ERROR;	/* error */
		}
	}

	/* load the new parameters */
	cpu->cpu_time = *c;
	cpu->cpu_period = *t;
	cpu->cpu_capacity = capacity;

	cpu_reserve_current_capacity = new_capacity;

	/* calculate cpu ticks per capacity */
	nanosec2tick(&qt, &cpu->cpu_period_ticks);
	nanosec2tick(&qc, &cpu->cpu_time_ticks);

	return RK_SUCCESS;	/* success */
}

static int
cpu_reserve_destroy(rk_reserve_t rsv)
{
	cpu_reserve_t	cpu = rsv->rsv_rsv;

#ifdef	DEBUG_RK
	printk("cpu_reserve_destroy: rsv(0x%x) cpu(0x%x)\n",
	       (int)rsv, (int)cpu);
#endif

#ifdef	UPERF_RK
	printk("start_account:%lu\n",
	       perf_average(result_start_account, counter_start_account));
	printk("start_account_timer:%lu\n",
	       perf_average(result_start_account_timer, counter_start_account_timer));
	printk("stop_account:%lu\n",
	       perf_average(result_stop_account, counter_stop_account));
	printk("replenish:%lu\n",
	       perf_average(result_replenish, counter_replenish));
	printk("enforce:%lu\n",
	       perf_average(result_enforce, counter_enforce));
#endif	/* UPERF_RK */

	/* destroy timer */
	rk_replenish_timer_destroy(rsv);

	/* return capacity */
	cpu_reserve_current_capacity -= cpu->cpu_capacity;
	if ((int)cpu_reserve_current_capacity < 0) {
		cpu_reserve_current_capacity = 0;
	}

#ifdef	EVAL_RK
	cpu_reserve_sched_timeshare_self();
	{
		unsigned long	i;
		struct	cpu_eval_data	*datap = cpu->cpu_eval_datap;
		/* dump data */
		printk("cpu_reserve: rsv(0x%x) param(%d/%d/%d) c/t(%lu/%lu)\n",
		       (int)rsv,
		       rsv->rsv_reserve_param.sch_mode,
		       rsv->rsv_reserve_param.enf_mode,
		       rsv->rsv_reserve_param.rep_mode,
		       (unsigned long)cpu->cpu_time_ticks,
		       (unsigned long)cpu->cpu_period_ticks);
		for (i = 0; i < cpu->cpu_eval_data_count; i++) {
			if (datap[i].c_used) {
				printk("cpu_reserve: rsv(0x%x) jiffies(%lu) used(%lu)\n",
				       (int)rsv, datap[i].t_jiffies, datap[i].c_used);
			}
			if (!(i%10)) {
				current->state = TASK_INTERRUPTIBLE;
				schedule_timeout(5);	/* sleep for 5 jiffies */
				schedule();
			}
		}
		/* free it */
		kfree(cpu->cpu_eval_datap);
	}
#endif

	list_del(&cpu->cpu_link);
	INIT_LIST_HEAD(&cpu->cpu_link);		/* for sure */
	kfree(cpu);
	rsv->rsv_rsv = NULL_RESERVE;

	rk_reserve_destroy(rsv);
	return RK_SUCCESS;	/* success */
}

/*
 *
 */
static inline void
cpu_reserve_summation_of_used_ticks(cpu_reserve_t cpu, cpu_tick_t now)
{
	cpu->cpu_period_used_ticks += (*now - cpu->cpu_period_start_ticks);
	cpu->cpu_period_start_ticks = *now;
}

/*
 * Replenish & Enforce a reserve
 */
static void
cpu_reserve_replenish(rk_reserve_t rsv, cpu_tick_t period)
{
	cpu_reserve_t	cpu = rsv->rsv_rsv;
	cpu_capacity_t	c;
	long long	ticks;

	*period = cpu->cpu_period_ticks;

	if (rsv->rsv_state & RSV_IS_RUNNING) {
		cpu_tick_data_t	now;
		rk_rdtsc(&now);
		cpu_reserve_summation_of_used_ticks(cpu, &now);
	}

	/* update statistics */
	cpu->cpu_period_prev_used_ticks = cpu->cpu_period_used_ticks;
	c = CAPACITY_OF(cpu->cpu_period_used_ticks, cpu->cpu_period_ticks);
	if (cpu->cpu_max_utilization < c) {
		cpu->cpu_max_utilization = c;
	} else if (cpu->cpu_min_utilization > c || cpu->cpu_min_utilization == 0) {
		cpu->cpu_min_utilization = c;
	}
	/* summation for average */
	cpu->cpu_average.total_utils += c;
	cpu->cpu_average.total_count++;

#if	defined(EVAL_RK)
	{
		if (cpu->cpu_eval_data_count < MAX_COUNT) {
			unsigned long		count = (cpu->cpu_eval_data_count)++;
			struct	cpu_eval_data	*datap = cpu->cpu_eval_datap;
			datap[count].t_jiffies = jiffies;
			datap[count].c_used = (unsigned long)cpu->cpu_period_used_ticks;
		}
	}
#endif
#if	defined(DEBUG_RK)
	printk("cpu_reserve_replenish: rsv(0x%x) used(%lu/%lu)\n",
	       (int)rsv, 
	       (unsigned long)cpu->cpu_period_used_ticks,
	       (unsigned long)cpu->cpu_period_available_ticks);
#endif

	/* refill the capacity */
	switch (rsv->rsv_reserve_param.rep_mode) {
	case RSV_SOFT:
		cpu->cpu_period_available_ticks = cpu->cpu_time_ticks;
		break;
	case RSV_FIRM:
		ticks = cpu->cpu_period_used_ticks - cpu->cpu_time_ticks;
		/* when excessively used */
		if (ticks > 0) {
			cpu_tick_data_t	next;
			next = cpu->cpu_time_ticks - ticks;
			if (next >= 0) {
				cpu->cpu_period_available_ticks = next;
			} else {
				cpu->cpu_period_available_ticks = 0;
			}
		} else {
			cpu->cpu_period_available_ticks = cpu->cpu_time_ticks;
		}
		break;
	case RSV_HARD:
		ticks = cpu->cpu_period_used_ticks - cpu->cpu_period_available_ticks;
		/* when excessively used */
		if (ticks > 0) {
			cpu_tick_data_t	next;
			next = cpu->cpu_time_ticks - ticks;
			if (next >= 0) {
				cpu->cpu_period_available_ticks = next;
			} else {
				cpu->cpu_period_available_ticks = 0;
			}
		} else {
			cpu->cpu_period_available_ticks = cpu->cpu_time_ticks;
		}
		break;
	}

	/* reset the used ticks */
	cpu->cpu_period_used_ticks = 0;


#ifdef	DEBUG_RK
	printk("cpu_reserve_replenish: rsv(0x%x) state(%x) next available(%lu)\n",
	       (int)rsv, 
	       (int)rsv->rsv_state,
	       (unsigned long)cpu->cpu_period_available_ticks);
#endif

	if (rsv->rsv_state & RSV_IS_DEPLETED) {
		switch (rsv->rsv_reserve_param.enf_mode) {
		/* when soft or firm enforcement, make it realtime */
		case RSV_SOFT:
		case RSV_FIRM:
			rsv->rsv_state &= ~RSV_IS_DEPLETED;
			rs_proc_list_apply(&rsv->rsv_rs->rs_proc_list,
					   cpu_reserve_sched_realtime);
			break;
		/* resume attached processes if this reserve is depleted and 
		 * there are available ticks for the next period */
		case RSV_HARD:
			if (cpu->cpu_period_available_ticks) {
				rsv->rsv_state &= ~RSV_IS_DEPLETED;
				wake_up(&cpu->depleted_wait);
			}
			break;
		}
	}
}

static void
cpu_reserve_enforce(rk_reserve_t rsv)
{
	if (!(rsv->rsv_state & RSV_IS_DEPLETED)) {
#ifdef	DEBUG_RK
#ifdef	DEBUG_RK_ENFORCE
		debug_rk_enforce_flag = 1;
#endif
		printk("cpu_reserve_enforce: rsv(0x%x)\n", (int)rsv);
#endif
		rsv->rsv_state |= RSV_IS_DEPLETED;
	}

	/* when soft or firm enforcement, make it timeshare */
#if 0
	if (rsv->rsv_reserve_param.enf_mode & (RSV_SOFT|RSV_FIRM)) {
		rs_proc_list_apply(&rsv->rsv_rs->rs_proc_list,
				   cpu_reserve_sched_timeshare);
	}
#endif
}

static void
cpu_reserve_sleep_on(rk_reserve_t rsv)
{
	cpu_reserve_t	cpu = rsv->rsv_rsv;
	sleep_on(&cpu->depleted_wait);
}

/*
 * Start & Stop the account of CPU utilization.
 */
static void
cpu_reserve_start_account(rk_reserve_t rsv)
{
	cpu_reserve_t	cpu = rsv->rsv_rsv;
	long long	next;

	next = cpu->cpu_period_available_ticks - cpu->cpu_period_used_ticks;

	switch (rsv->rsv_state) {
	case RSV_IS_NULL:
		rsv->rsv_state |= RSV_IS_RUNNING;
		if (next > 0) {
			rk_enforce_timer_start(rsv, &next,
					       &cpu->cpu_period_start_ticks);
		} else {
			rsv->rsv_state |= RSV_IS_DEPLETED;
			rk_rdtsc(&cpu->cpu_period_start_ticks);
		}
		break;
	case RSV_IS_DEPLETED:
		rsv->rsv_state |= RSV_IS_RUNNING;
		rk_rdtsc(&cpu->cpu_period_start_ticks);
		break;
	case RSV_IS_RUNNING:			/* must not be so */
		printk("cpu_reserve_start_account: rsv(0x%x) illegal state(RUNNING)\n",
		       (int)rsv);
		if (next > 0) {
			rk_enforce_timer_start(rsv, &next,
					       &cpu->cpu_period_start_ticks);
		} else {
			rsv->rsv_state |= RSV_IS_DEPLETED;
			rk_rdtsc(&cpu->cpu_period_start_ticks);
		}
		break;
	case RSV_IS_RUNNING|RSV_IS_DEPLETED:	/* must not be so */
		printk("cpu_reserve_start_account: rsv(0x%x) illegal state(RUNNING|DEPLETED)\n",
		       (int)rsv);
		break;
	default:		/* unkown */
		printk("cpu_reserve_start_account: rsv(0x%x) unkown state(0x%x) next available(%lu)\n",
		       (int)rsv, (int)rsv->rsv_state, (unsigned long)next);
		rsv->rsv_state = RSV_IS_RUNNING;
		if (next > 0) {
			rk_enforce_timer_start(rsv, &next,
					       &cpu->cpu_period_start_ticks);
		} else {
			rsv->rsv_state |= RSV_IS_DEPLETED;
			rk_rdtsc(&cpu->cpu_period_start_ticks);
		}
		break;
	}
}

static inline void
cpu_reserve_check_enforce(rk_reserve_t rsv, cpu_reserve_t cpu)
{
	if (cpu->cpu_period_used_ticks >= cpu->cpu_period_available_ticks) {
#ifdef	DEBUG_RK
		printk("cpu_reserve_check_enforce: enforce rsv(0x%x) used(%lu/%lu)\n",
		       (int)rsv,
		       (unsigned long)cpu->cpu_period_used_ticks,
		       (unsigned long)cpu->cpu_period_available_ticks);
#endif
		/* enforcing a reserve */
		cpu_reserve_enforce(rsv);
	}
}

static void
cpu_reserve_stop_account(rk_reserve_t rsv, cpu_tick_t now)
{
	cpu_reserve_t	cpu = rsv->rsv_rsv;

	cpu_reserve_summation_of_used_ticks(cpu, now);
	
	switch (rsv->rsv_state) {
	case RSV_IS_RUNNING:
		rsv->rsv_state &= ~RSV_IS_RUNNING;
		cpu_reserve_check_enforce(rsv, cpu);
		break;
	case RSV_IS_RUNNING|RSV_IS_DEPLETED:
		rsv->rsv_state &= ~RSV_IS_RUNNING;
		break;
	case RSV_IS_NULL:	/* must not be so */
		printk("cpu_reserve_stop_account: rsv(0x%x) illegal state(NULL)\n",
		       (int)rsv);
		cpu_reserve_check_enforce(rsv, cpu);
		break;
	case RSV_IS_DEPLETED:	/* must not be so */
		printk("cpu_reserve_stop_account: rsv(0x%x) illegal state(NULL|DEPLETED)\n",
		       (int)rsv);
		break;
	default:		/* unkown */
		printk("cpu_reserve_stop_account: rsv(0x%x) unkown state(0x%x))\n",
		       (int)rsv, (int)rsv->rsv_state);
		rsv->rsv_state = RSV_IS_NULL;
		cpu_reserve_check_enforce(rsv, cpu);
		break;
	}

}

static	void
cpu_reserve_attach(rk_reserve_t rsv, struct rs_proc_list* rs_proc)
{
	if (rsv->rsv_state & RSV_IS_DEPLETED) {
		/* should stop it? */
		return;
	}
	cpu_reserve_sched_realtime(rs_proc);
}

static	void
cpu_reserve_detach(rk_reserve_t rsv, struct rs_proc_list* rs_proc)
{
	cpu_reserve_sched_timeshare(rs_proc);
}

/*
 * Linux specific
 */
#ifdef	linux
/*
 * Linux proc file system interface
 */
#include <asm/uaccess.h>
asmlinkage rk_reserve_t
sys_cpu_reserve_create(rk_resource_set_t rs,
		       struct timespec *uc, struct timespec *ut,
		       struct rk_reserve_param *up)
{
	struct timespec c, t;
	struct rk_reserve_param p;

	copy_from_user(&c, uc, sizeof(struct timespec));
	copy_from_user(&t, ut, sizeof(struct timespec));
	copy_from_user(&p, up, sizeof(struct rk_reserve_param));

	return cpu_reserve_create(rs, &c, &t, &p);
}

asmlinkage int
sys_cpu_reserve_ctl(rk_resource_set_t rs, rk_reserve_t rsv,
		    struct timespec *uc, struct timespec *ut)
{
	struct timespec c, t;

	copy_from_user(&c, uc, sizeof(struct timespec));
	copy_from_user(&t, ut, sizeof(struct timespec));

	return cpu_reserve_ctl(rs, rsv, &c, &t);
}

#define	LINUX_REALTIME_PRIORITY		50
#define	LINUX_TIMESHARE_PRIORITY	0

static inline void
cpu_reserve_setscheduler(struct task_struct *p,
			 unsigned long policy, unsigned long rt_priority)
{
	p->policy = policy;
	p->rt_priority = rt_priority;
	current->need_resched = 1;
}

void
cpu_reserve_sched_realtime(struct rs_proc_list* rs_proc)
{
	struct task_struct *p = rs_proc->rs_proc_task;
	unsigned long flags;

	if (!(p->policy & SCHED_FIFO)) {
#ifdef	DEBUG_RK
		printk("cpu_reserve_sched_realtime: pid(%d)\n", p->pid);
#endif
		rk_spin_lock(flags);
		{
			cpu_reserve_setscheduler(p, SCHED_FIFO, LINUX_REALTIME_PRIORITY);
		}
		rk_spin_unlock(flags);
	}
}

void
cpu_reserve_sched_timeshare(struct rs_proc_list* rs_proc)
{
	struct task_struct *p = rs_proc->rs_proc_task;
	unsigned long flags;

	if (p->policy & SCHED_FIFO) {
#ifdef	DEBUG_RK
		printk("cpu_reserve_sched_timeshare: pid(%d)\n", p->pid);
#endif
		rk_spin_lock(flags);
		{
			cpu_reserve_setscheduler(p, SCHED_OTHER, LINUX_TIMESHARE_PRIORITY);
		}
		rk_spin_unlock(flags);
	}
}

void
cpu_reserve_sched_timeshare_self(void)
{
	struct task_struct *p = current;
	unsigned long flags;

	if (p->policy & SCHED_FIFO) {
#ifdef	DEBUG_RK
		printk("cpu_reserve_sched_timeshare_self: pid(%d)\n", p->pid);
#endif
		rk_spin_lock(flags);
		{
			cpu_reserve_setscheduler(p, SCHED_OTHER, LINUX_TIMESHARE_PRIORITY);
		}
		rk_spin_unlock(flags);
	}
}

static int
cpu_reserve_read_proc(rk_reserve_t rsv, char *buf)
{
	char		*p = buf;
	cpu_reserve_t	cpu = rsv->rsv_rsv;
	unsigned long	used_cpu_time;
	cpu_capacity_t	ave, cur, prv;

	used_cpu_time = TICK2USEC(&cpu->cpu_period_used_ticks);
	cur = CAPACITY_OF(cpu->cpu_period_used_ticks, cpu->cpu_period_ticks);
	prv = CAPACITY_OF(cpu->cpu_period_prev_used_ticks, cpu->cpu_period_ticks);

	if (cpu->cpu_average.total_count) {
		ave = cpu->cpu_average.total_utils/cpu->cpu_average.total_count;
	} else {
		ave = cur;
	}

	/* 1st line: enf_mode, rep_mode */
	p += sprintf(p, "%d %d\n",
		     (int)rsv->rsv_reserve_param.enf_mode,
		     (int)rsv->rsv_reserve_param.rep_mode);

	/* 2nd line: utilization statistics
	 * reseved, previous, average, max, min */
	p += sprintf(p, "%lu.%02lu %lu.%02lu %lu.%02lu %lu.%02lu %lu.%02lu\n",
		     CAPACITY_INT(cpu->cpu_capacity), CAPACITY_FRAC(cpu->cpu_capacity),
		     CAPACITY_INT(prv), CAPACITY_FRAC(prv),
		     CAPACITY_INT(ave), CAPACITY_FRAC(ave),
		     CAPACITY_INT(cpu->cpu_max_utilization),
			CAPACITY_FRAC(cpu->cpu_max_utilization),
		     CAPACITY_INT(cpu->cpu_min_utilization),
			CAPACITY_FRAC(cpu->cpu_min_utilization));

	/* 3rd line: current CPU usage 
	 * used available requested */
	p += sprintf(p, "%lu %lu %lu\n",
		     used_cpu_time, 
		     TICK2USEC(&cpu->cpu_period_available_ticks),
		     TICK2USEC(&cpu->cpu_time_ticks));
	/* 4th line: count of replenishment */
	p += sprintf(p, "%lu\n", cpu->cpu_average.total_count);

#ifdef	UPERF_RK
	p += sprintf(p, "start_account:%lu\n",
		     perf_average(result_start_account, counter_start_account));
	p += sprintf(p, "stop_account:%lu\n",
		     perf_average(result_stop_account, counter_stop_account));
	p += sprintf(p, "replenish:%lu\n",
		     perf_average(result_replenish, counter_replenish));
	p += sprintf(p, "enforce:%lu\n",
		     perf_average(result_enforce, counter_enforce));
#endif

	return (p - buf);
}

int
cpu_reserve_status_proc(char *buf)
{
	char		*p = buf;

	/* cpu ticks per second, CPU capacity taken for reserves */
	p += sprintf(p, "%lu %lu.%02lu\n",
		     (unsigned long)rk_cpu_ticks_per_second,
		     CAPACITY_INT(cpu_reserve_current_capacity),
		     CAPACITY_FRAC(cpu_reserve_current_capacity));

	return (p - buf);
}
#endif	/* linux */
