/*
 * Road map of net reserve:
 *
 *	1. create a net reserve and counting a total number of bytes sent through it
 *	2. perform accounting based on C/T model
 *	3. introduce the enforcement for net reserves
 */
#include <rk/rk_linux.h>
#include <rk/rk_error.h>
#include <rk/rk.h>
#include <linux/time.h>
#include <asm/processor.h>

#ifdef	RK_NET_RSV

static	int		net_reserve_destroy(rk_reserve_t);
static	void		net_reserve_replenish(rk_reserve_t, cpu_tick_t);
static	void		net_reserve_enforce(rk_reserve_t);
static	void		net_reserve_attach(rk_reserve_t, struct rs_proc_list*);
static	void		net_reserve_detach(rk_reserve_t, struct rs_proc_list*);
static	void		net_reserve_sleep_on(rk_reserve_t);
static	int		net_reserve_read_proc(rk_reserve_t, char *);

struct rk_reserve_ops	net_reserve_ops = {
	net_reserve_destroy,
	NULL,				/* start_account */
	NULL,				/* stop_account */
	net_reserve_replenish,
	net_reserve_enforce,
	net_reserve_attach,
	net_reserve_detach,
#ifdef	linux
	net_reserve_sleep_on,
	net_reserve_read_proc,
#endif
};

/*
 * Internal data structure to keep track of cpu capacity usage
 *	99.99% is represented as 9999
 */
typedef	unsigned long		net_capacity_t;
typedef	unsigned long long	net_capacity_quad_t;

#define	CAPACITY_INT(x)		((x)/100)
#define	CAPACITY_FRAC(x)	((x)-(x/100)*100)
#define	INT2CAPACITY(x)		((x)*10000)
#define	CAPACITY2INT(x)		((x)/10000)
#define	PERCENT2CAPACITY(x)	((x)*100)
#define	CAPACITY_OF(c,t)	(INT2CAPACITY(c)/(t))

/*
 *	66% ... is the maximum available capacity for reserves.
 */
#define	NET_CAPACITY_MAX	PERCENT2CAPACITY(66)
#define	LINK_SPEED_10MBPS	(10*1000*1000/8)

/*
 * net_reserve structure
 */
struct net_reserve {
	/*current period */
	cpu_tick_data_t		start_ticks;		/* remember start ticks for 
							   the current invocation */
	size_t			used_amount;		/* used amount
							   in the current period */
	size_t			available_amount;	/* available amount
							   for this period */
	struct wait_queue 	*depleted_wait;		/* for enforcement */
	/* classifier & queueing discipline */
							/* sk_buff queue head */
	/* execution statistics */
	net_capacity_t		max_utilization;
	net_capacity_t		min_utilization;
	struct {
		unsigned long		total_count;	/* total count of replenishment */
		net_capacity_quad_t	total_utils;	/* summation of utilizations */
	}			average;
	/* static data */
	struct list_head	net_link;		/* link all net reserves */
	cpu_tick_data_t		period_ticks;		/* CPU ticks for period */
	net_capacity_t		capacity;		/* requested capacity */
	size_t			amount;			/* B/W per period */
	struct timespec		period;
#ifdef	linux
	/* <socket, device> */
	struct sock		*sk;
	struct device		*netdev;
#endif	/* linux */
};
typedef	struct net_reserve	*net_reserve_t;

/*
 * Capacity must be administrated for each interface
 */
struct list_head	net_reserve_head;
net_capacity_t		net_reserve_current_capacity;	/* for admission control */

/*
 *
 */
void
net_reserve_init(void)
{
#ifdef	linux
	extern struct Qdisc	*net_reserve_qdisc;
#endif

	INIT_LIST_HEAD(&net_reserve_head);
	net_reserve_current_capacity = 0;
	net_reserve_qdisc = NULL;
}

rk_reserve_t
net_reserve_create(rk_resource_set_t rs,
		   size_t b, struct timespec *t,
		   rk_reserve_param_t p,
		   void *data)
{
	rk_reserve_t	rsv;
	net_reserve_t	netrsv;
	net_capacity_t	capacity;
	net_capacity_quad_t	qc, qt;
#ifdef	linux
	struct device	*netdev = data;
#endif

	/* admission control */
	qc = b;
	qt = t->tv_sec; qt *= NANOSEC; qt += t->tv_nsec;
	capacity = CAPACITY_OF(qc, (qt*LINK_SPEED_10MBPS)/NANOSEC);

	/* create net reserve object and set static data */
	netrsv = malloc(sizeof(struct net_reserve));
	bzero(netrsv, sizeof(struct net_reserve));
	netrsv->period = *t;
	netrsv->amount = b;
	nanosec2tick(&qt, &netrsv->period_ticks);
	netrsv->capacity = capacity;
	list_add(&netrsv->net_link, &net_reserve_head);

	/* init per period data */
	netrsv->available_amount = netrsv->amount;

	/* init waitqueue */
	init_waitqueue(&netrsv->depleted_wait);

	/* create generic reserve object */
	rsv = rk_reserve_create(rs, RSV_NET);
	rsv->rsv_state = RSV_IS_NULL;
	rsv->rsv_rsv = netrsv;
	rsv->rsv_ops = &net_reserve_ops;
	rsv->rsv_reserve_param = *p;

	/* create a timer for it */
	rk_replenish_timer_create(rsv, &netrsv->period_ticks);

	/* attach reserve to resource set*/
	rk_resource_set_attach_reserve(rs, rsv);

	return rsv;		/* success */
}

static int
net_reserve_ctl(rk_resource_set_t rs, rk_reserve_t rsv,
		size_t b, struct timespec *t)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;

	return RK_SUCCESS;	/* success */
}

/*
 *
 */
static int
net_reserve_destroy(rk_reserve_t rsv)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;

	/* destroy timer */
	rk_replenish_timer_destroy(rsv);

	/* return capacity */
	net_reserve_current_capacity -= netrsv->capacity;
	if ((int)net_reserve_current_capacity < 0) {
		net_reserve_current_capacity = 0;
	}

	list_del(&netrsv->net_link);
	INIT_LIST_HEAD(&netrsv->net_link);	/* for sure */
	kfree(netrsv);
	rsv->rsv_rsv = NULL_RESERVE;

	rk_reserve_destroy(rsv);
	return RK_SUCCESS;	/* success */
}

static void
net_reserve_replenish(rk_reserve_t rsv, cpu_tick_t period)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;
}

static void
net_reserve_enforce(rk_reserve_t rsv)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;
}


static	void
net_reserve_attach(rk_reserve_t rsv, struct rs_proc_list* rs_proc)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;
}

static	void
net_reserve_detach(rk_reserve_t rsv, struct rs_proc_list* rs_proc)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;
}

static void
net_reserve_sleep_on(rk_reserve_t rsv)
{
	net_reserve_t	netrsv = rsv->rsv_rsv;
	sleep_on(&netrsv->depleted_wait);
}

/*
 * Linux specific
 */
#ifdef	linux
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/init.h>
#include <net/sock.h>
#include <net/pkt_sched.h>
/*
 * Linux Packet Scheduler
 *
 *	enqueue: classifies a packet (sk_buff) 
 *		 and puts it in an appropriate queue in qdisc
 *	dequeue: removes the highest priority packet from a queue
 *	requeue: enqueues a packet again (when its transmission is failed)
 *		 always succeeds (enqueue may fail because of too long queue)
 *	init:	 initializes queues
 *	reset:	 removes all packets from queues and clears everything
 *
 */

/*
 *
 */
struct Qdisc	*net_reserve_qdisc;

/* 
 * Net-reserve extends "3-band FIFO queue" in sch_generic.c by adding a 
 * RT queue.  There must be a RT queue for each priority level, but there is
 * only a single queue now to make it different from best effort queues.
 */

static const u8 prio2band[TC_PRIO_MAX+1] =
	{ 2, 3, 3, 3, 2, 3, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2 };
/* was	{ 1, 2, 2, 2, 1, 2, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1 };*/

static int
net_reserve_enqueue(struct sk_buff *skb, struct Qdisc* qdisc)
{
	struct sk_buff_head *list;
	struct sock *sk = skb->sk;
	struct net_reserve *net_rsv;
	

	/* if sk is associated with a net reserve */
	if (sk->rk_net_reserve) {
		/* obtain a real reserve */
		net_rsv = sk->rk_net_reserve->rsv_rsv;

		/* account for the utilization */
		net_rsv->used_amount += skb->len;	/* add the size of a packet */
#if 0
		if (net_rsv->used_amount > net_rsv->available_amount) {
		}
#endif

		/* choose RT queue */
		list = ((struct sk_buff_head*)qdisc->data);

		/* enqueue now */
		__skb_queue_tail(list, skb);
		qdisc->q.qlen++;
		return 1;
	}

	/* No net reserve */
	list = ((struct sk_buff_head*)qdisc->data) +
		prio2band[skb->priority&TC_PRIO_MAX];

	if (list->qlen <= skb->dev->tx_queue_len) {
		__skb_queue_tail(list, skb);
		qdisc->q.qlen++;
		return 1;
	}
	qdisc->stats.drops++;
	kfree_skb(skb);
	return 0;
}

static struct sk_buff *
net_reserve_dequeue(struct Qdisc* qdisc)
{
	int prio;
	struct sk_buff_head *list = ((struct sk_buff_head*)qdisc->data);
	struct sk_buff *skb;

	for (prio = 0; prio < 4; prio++, list++) {
		skb = __skb_dequeue(list);
		if (skb) {
			qdisc->q.qlen--;
			return skb;
		}
	}
	return NULL;
}


static int
net_reserve_requeue(struct sk_buff *skb, struct Qdisc* qdisc)
{
	struct sk_buff_head *list;
	struct sock *sk = skb->sk;

	/* if sk is associated with a net reserve */
	if (sk->rk_net_reserve) {
		/* choose RT queue */
		list = ((struct sk_buff_head*)qdisc->data);
	} else {
		list = ((struct sk_buff_head*)qdisc->data) +
			prio2band[skb->priority&TC_PRIO_MAX];
	}

	__skb_queue_head(list, skb);
	qdisc->q.qlen++;
	return 1;
}

static int 
net_reserve_queue_init(struct Qdisc *qdisc, struct rtattr *opt)
{
	int i;
	struct sk_buff_head *list;

	list = ((struct sk_buff_head*)qdisc->data);

	for (i=0; i < 4; i++)
		skb_queue_head_init(list+i);

	return 0;
}

static void
net_reserve_queue_reset(struct Qdisc* qdisc)
{
	int prio;
	struct sk_buff_head *list = ((struct sk_buff_head*)qdisc->data);

	for (prio=0; prio < 4; prio++)
		skb_queue_purge(list+prio);
	qdisc->q.qlen = 0;
}

static struct Qdisc_ops net_reserve_queue_ops =
{
	NULL,
	NULL,
	"net_reserve",		/* must be less than IFNAMSIZ (16) characters */
	4 * sizeof(struct sk_buff_head),
				/* 3-band FIFO queue + 1 for RT (for now) */
	net_reserve_enqueue,
	net_reserve_dequeue,
	net_reserve_requeue,
	NULL,

	net_reserve_queue_init,
	net_reserve_queue_reset,
};

void
net_reserve_activate(struct device *dev)
{
	if (net_reserve_qdisc == NULL) {
		net_reserve_qdisc = qdisc_create_dflt(dev, &net_reserve_queue_ops);
	}

	if (net_reserve_qdisc->dev != dev) {
		printk("net_reserve_activate: multiple devices not supported (yet).\n");
	}

	dev_set_scheduler(dev, net_reserve_qdisc);
}

/*
 *	struct sock *sk;		... include/net/sock.h
 *	struct socket *sock;		... include/linux/net.h
 *	struct sk_buff *skb;		... include/linux/skbuff.h
 *	struct device *netdev		... include/linux/netdevice.h
 *
 *	sock = sk->socket;
 *	sk = sock->sk;
 *	sk = skb->sk;
 *	netdev = skb->dev;
 *
 *	packet classifier uses "struct sock"
 */

asmlinkage rk_reserve_t
sys_net_reserve_create(rk_resource_set_t rs,
		       size_t ub, struct timespec *ut,
		       struct rk_reserve_param *up,
		       int socket_fd, struct ifreq *uifr)
{
	rk_reserve_t	rsv;
	net_reserve_t	netrsv;
	struct timespec	t;
	struct rk_reserve_param p;
	struct socket	*sock;
	struct sock	*sk = sock->sk;
	struct device	*netdev;
	struct ifreq	ifr;
	int	err;
	extern struct socket *sockfd_lookup(int, int*);

	copy_from_user(&t, ut, sizeof(struct timespec));
	copy_from_user(&p, up, sizeof(struct rk_reserve_param));
	copy_from_user(&ifr, uifr, sizeof(struct ifreq));

	/* correct file descriptor for a socket? */
	sock = sockfd_lookup(socket_fd, &err);
	if (sock == NULL) {
		printk("sys_net_reserve_create: invalid socket file descriptor(%d)\n",
		       socket_fd);
		return NULL_RESERVE;
	}
	/* correct device name? */
	netdev = dev_get(ifr.ifr_name);
	if (netdev == NULL) {
		printk("sys_net_reserve_create: invalid device name(%s)\n",
		       ifr.ifr_name);
		return NULL_RESERVE;
	}

	rsv = net_reserve_create(rs, ub, &t, &p, netdev);
	if (rsv) {
		sk = sock->sk;
		sk->rk_net_reserve = rsv;
		/* */
		netrsv = rsv->rsv_rsv;
		netrsv->sk = sk;
		netrsv->netdev = netdev;
	}

	/* net_reserve_qdisc active ? */
	if (net_reserve_qdisc == NULL) {
		/* activate net reserve Qdisc */
		net_reserve_activate(netdev);
	}
	
	return rsv;
}

asmlinkage int
sys_net_reserve_ctl(rk_resource_set_t rs, rk_reserve_t rsv,
		    size_t ub, struct timespec *ut)
{
	struct timespec t;

	copy_from_user(&t, ut, sizeof(struct timespec));

	return net_reserve_ctl(rs, rsv, ub, &t);
}

/*
 * Linux proc file system interface
 */

static int
net_reserve_read_proc(rk_reserve_t rsv, char *buf)
{
	char		*p = buf;
	net_reserve_t	netrsv = rsv->rsv_rsv;
	net_capacity_t	ave;

	if (netrsv->average.total_count) {
		ave = netrsv->average.total_utils/netrsv->average.total_count;
	} else {
		if (netrsv->available_amount) {
			ave = CAPACITY_OF(netrsv->used_amount, netrsv->available_amount);
		}
		else {
			ave = CAPACITY_OF(netrsv->used_amount, 1);
		}
	}

	p += sprintf(p, "utilization (ave max min): %lu.%02lu %lu.%02lu %lu.%02lu\n",
		     CAPACITY_INT(ave), CAPACITY_FRAC(ave),
		     CAPACITY_INT(netrsv->max_utilization),
		     	CAPACITY_FRAC(netrsv->max_utilization),
		     CAPACITY_INT(netrsv->min_utilization),
		     	CAPACITY_FRAC(netrsv->min_utilization));
	p += sprintf(p, "used: %d ()/%d %d\n",
		     netrsv->used_amount,
		     netrsv->available_amount,
		     netrsv->amount);
	p += sprintf(p, "count: %lu\n", netrsv->average.total_count);
	p += sprintf(p, "reserved: %lu.%02lu\n",
		     CAPACITY_INT(netrsv->capacity), CAPACITY_FRAC(netrsv->capacity));

	return (p - buf);
}

int
net_reserve_status_proc(char *buf)
{
	char		*p = buf;

	p += sprintf(p, "net_reserve_current_capacity: %lu.%02lu\n",
		     CAPACITY_INT(net_reserve_current_capacity),
		     CAPACITY_FRAC(net_reserve_current_capacity));

	return (p - buf);
}
#endif	/* linux */

#else	/* RK_NET_RSV */
#ifdef	linux
asmlinkage rk_reserve_t
sys_net_reserve_create(rk_resource_set_t rs,
		       size_t ub, struct timespec *ut,
		       struct rk_reserve_param *up,
		       int socket_fd, void *uifr)
{
	return NULL;
}

asmlinkage int
sys_net_reserve_ctl(rk_resource_set_t rs, rk_reserve_t rsv,
		    size_t ub, struct timespec *ut)
{
	return RK_ERROR;
}

#endif	/* linux */
#endif	/* RK_NET_RSV */
