/*
 * RK: resource set management
 */
#include <rk/rk_linux.h>
#include <rk/rk_error.h>
#include <rk/rk.h>

rk_resource_set_t	rk_current_resource_set;
LIST_HEAD(rk_resource_set_root);

void
rk_resource_set_init(void)
{
	rk_current_resource_set = NULL_RESOURCE_SET;
	INIT_LIST_HEAD(&rk_resource_set_root);
}

rk_resource_set_t
rk_resource_set_create(void)
{
	rk_resource_set_t	rs;

	rs = malloc(sizeof(struct rk_resource_set));
	bzero(rs, sizeof(struct rk_resource_set));

	if (list_empty(&rk_resource_set_root)) {
		/* enable rk scheduling hook */
		rk_enable();
	}

	list_add(&rs->rs_rs, &rk_resource_set_root);

	rs->rs_cpu = NULL_RESERVE;

	INIT_LIST_HEAD(&rs->rs_devices);

#ifdef	linux
	INIT_LIST_HEAD(&rs->rs_proc_list);	/* just for sure */
	rk_proc_rs_create(rs);			/* create directory as /proc/rk/<rs_id>/ */
#endif

	return rs;
}

/*
 * rk_resource_set_destroy_reserves(struct list_head *rsv_list)
 *	called from rk_resource_set_destroy to destroy all the reserves 
 *	on the list specified by ``rsv_list''.
 */
static void
rk_resource_set_destroy_reserves(struct list_head *rsv_list)
{
	rk_reserve_t	rsv;
#ifdef DEBUG_RK
	printk("rk_resource_set_destroy_reserves: rsv_list(0x%x)\n",
	       (unsigned int)rsv_list);
#endif
	while (!list_empty(rsv_list)) {
		/* get a reserve */
		rsv = list_entry(rsv_list->next, struct rk_reserve, rsv_rs_link);
		list_del(&rsv->rsv_rs_link);
		/* detroys all */
#ifdef DEBUG_RK
		printk("rk_resource_set_destroy_reserves: rsv(0x%x)\n",
		       (unsigned int)rsv);
		printk("rk_resource_set_destroy_reserves: rsv_ops(0x%x) detroy(0x%x)\n",
		       (unsigned int)rsv->rsv_ops,
		       (unsigned int)rsv->rsv_ops->destroy);
#endif
		(rsv->rsv_ops->destroy)(rsv);
	}
}

int
rk_resource_set_destroy(rk_resource_set_t rs)
{
	rk_reserve_t	cpu = rs->rs_cpu;

#ifdef DEBUG_RK
	printk("rk_resource_set_destroy(0x%x)\n", (unsigned int)rs);
#endif

	if (list_empty(&rs->rs_rs)) {	/* more sanity check necessary */
		printk("rk_resource_set_destroy: rs(0x%x) not on list\n",
		       (unsigned int)rs);
		/* cannot destory non-active resource set */
		return RK_ERROR;	/* error */
	}

	/* stop it if this is the current */
	if (rk_current_resource_set == rs) {
		cpu_tick_data_t	now;
		rk_current_resource_set = NULL_RESOURCE_SET;
		rk_enforce_timer_cancel();
		rk_resource_set_stop_account(rs, &now);
	}

	/* start cleaning up */

	/* free attached processes/threads */
#ifdef	linux
	{
		static void rk_resource_set_detach_all_processes(rk_reserve_t, struct list_head *);
		rk_resource_set_detach_all_processes(cpu, &rs->rs_proc_list);
	}
#endif	/* linux */

	/* free attached reserves if there is any of them */
	if (cpu) {
		(cpu->rsv_ops->destroy)(cpu);
	}
	rk_resource_set_destroy_reserves(&rs->rs_devices);	/* device reserves? */

#ifdef	linux
	/* remove /proc/rk/<rs_id> entry
	 * shuld be after removing its contents */
	rk_proc_rs_destroy(rs);
#endif	/* linux */

	/* finally free resource set */
	list_del(&rs->rs_rs);
	INIT_LIST_HEAD(&rs->rs_rs);	/* just for sure */
	free(rs);
	if (list_empty(&rk_resource_set_root)) {
		/* disable rk scheduling hook */
		rk_disable();
	}

	return RK_SUCCESS;	/* success */
}

int
rk_resource_set_attach_reserve(rk_resource_set_t rs, rk_reserve_t rsv)
{

	if (!list_empty(&rsv->rsv_rs_link)) {	/* if already attached */
		return RK_ERROR;	/* error */
	}

	switch (rsv->rsv_type) {
	case RSV_CPU:		/* cpu capacity */
		/* more sanity check */
		if (rs->rs_cpu) {	/* resource set has cpu reserve */
			/* 
			 * at this moment, a resource set can use only 
			 * a single cpu reserve.  Multiple cpu reserves should
			 * be able to be utilized by a resource set.
			 */
			return RK_ERROR;	/* error */
		}
		/* */
		rs->rs_cpu = rsv;
		rsv->rsv_rs = rs;
		break;
	case RSV_DEV:		/* virtual devices */
		list_add(&rsv->rsv_rs_link, &rs->rs_devices);
		rsv->rsv_rs = rs;
		break;
	default:		/* unknown reserve type */
		break;
		return RK_ERROR;	/* error */
	}

	return RK_SUCCESS;	/* success */
}

void
rk_resource_set_detach_reserve(rk_resource_set_t rs, rk_reserve_t rsv)
{
	/* detach from rk_resource_set */
	list_del(&rsv->rsv_rs_link);

	/* clear link */
	rsv->rsv_rs = NULL_RESOURCE_SET;
	INIT_LIST_HEAD(&rsv->rsv_rs_link);
}

/*
 * Functions depending on the upper OS kernel
 *
 */
#ifdef	linux

asmlinkage rk_resource_set_t
sys_resource_set_create(void)
{
	return rk_resource_set_create();
}

asmlinkage int 
sys_resource_set_destroy(rk_resource_set_t rs)
{
	return rk_resource_set_destroy(rs);
}

asmlinkage int
sys_resource_set_attach_process(rk_resource_set_t rs, pid_t pid)
{
	rk_reserve_t		cpu = rs->rs_cpu;
	struct rs_proc_list	*rs_proc;
	struct task_struct	*tsk;

	tsk = __rk_find_process_by_pid(pid);
#ifdef	DEBUG_RK
	printk("resource_set_attach_process: rs(0x%x) pid(%d)->task(0x%x)\n",
	       (int)rs, pid, (int)tsk);
#endif

	if (!tsk) {
		printk("resource_set_attach_process: invalid pid(%d)\n",
		       pid);
		return RK_ERROR;
	}
	
	if (tsk->rk_resource_set) {
		printk("resource_set_attach_process: resource set already attached to this process. pid(%d) rs(0x%x)\n",
		       tsk->pid, (int)tsk->rk_resource_set);
		return RK_ERROR;
	}

	tsk->rk_resource_set = rs;

	rs_proc = malloc(sizeof(struct rs_proc_list));
	bzero(rs_proc, sizeof(struct rs_proc_list));

	rs_proc->rs_proc_task = tsk;
	rs_proc->rs_proc_pid = pid;

	list_add(&rs_proc->rs_proc_list, &rs->rs_proc_list);
#if 0
	/* enable rk scheduling hook */
	rk_enable();	/* move into "rk_resource_set_create() */
#endif
	/* */
	if (cpu) {
		cpu->rsv_ops->attach(cpu, rs_proc);
	}
	/* if this is the current process,
	 * make it the current resource set and start accouting now */
	if (current == tsk && rk_current_resource_set != rs) {
		rk_current_resource_set = rs;
		if (cpu) {
			cpu->rsv_ops->start_account(cpu);
			if (cpu->rsv_state & RSV_IS_DEPLETED) {
				cpu->rsv_ops->enforce(cpu);
			}
		}
	}
	return RK_SUCCESS;
}

static void
rk_resource_set_detach_all_processes(rk_reserve_t cpu, struct list_head *proc_list)
{
	struct rs_proc_list *rs_proc;

	while (!list_empty(proc_list)) {
		/* get a rs_proc */
		rs_proc = list_entry(proc_list->next,
				     struct rs_proc_list, rs_proc_list);
		list_del(&rs_proc->rs_proc_list);
		/* detach reserve */
		cpu->rsv_ops->detach(cpu, rs_proc);
		/* detach resource set from task */
		rs_proc->rs_proc_task->rk_resource_set = NULL_RESOURCE_SET;
		/* restore its original scheduling parameters */
		/* detroys it */
		INIT_LIST_HEAD(&rs_proc->rs_proc_list);	/* just for sure */
		free(rs_proc);
	}
}

int
resource_set_status_proc(char *buf)
{
	char		*p = buf;
	rk_resource_set_t	rs;
	struct list_head	*e;

	e = rk_resource_set_root.next;
	while(e != &rk_resource_set_root) {
		rs = list_entry(e, struct rk_resource_set, rs_rs);
		p += sprintf(p, "%x\n", (int)rs);
		e = e->next;
	}

	return (p - buf);
}
#endif
