/*
 * RK: timer management for reserves
 */
#include <rk/rk_linux.h>
#include <rk/rk_error.h>
#include <rk/rk.h>
#include <linux/timer.h>

#ifdef	DEBUG_RK
/**/
#define	DEBUG_RK_LINUX_TIMER
#undef	DEBUG_RK_LINUX_TIMER2
#define	DEBUG_RK_ENFORCE_TIMER
#define	DEBUG_RK_ENFORCE_TIMER_START
#undef	DEBUG_RK_ENFORCE_TIMER_ISR		/* rk_timer_isr */
/**/
#endif	/*DEBUG_RK*/

#define	HIGH_RESOLUTION_TIMER	1		/* 1 = on, 0 = off */

LIST_HEAD(rk_free_timer_root);			/* root of free timers */
LIST_HEAD(rk_online_timer_root);		/* root of on-line timers */

cpu_tick_data_t		rk_cpu_tick_at_boot;
cpu_tick_data_t		rk_cpu_tick_at_last_jiffy;

/*
 * Enforcement & replenishment are relied on different "OS dependent" timers
 * because of the differences of their natures.
 *
 *
 */
void
rk_timer_init(void)
{
}

/*
 * Replenishment timer managemnet functions
 */
rk_timer_t
rk_timer_create(void)
{
	rk_timer_t	tmr;

	if (list_empty(&rk_free_timer_root)) {
		/* allocate a new one */
		tmr = malloc(sizeof(struct rk_timer));
		bzero(tmr, sizeof(struct rk_timer));
	} else {
		unsigned long flags;
		rk_spin_lock(flags);
		{
			/* there is one in free list */
			tmr = list_entry(rk_free_timer_root.next, struct rk_timer, tmr_tmr);
			list_del(&tmr->tmr_tmr);
			/*bzero(tmr, sizeof(struct rk_timer));*/
		}
		rk_spin_unlock(flags);
	}

	return tmr;
}

static inline void
rk_timer_destroy(rk_timer_t tmr)
{
	unsigned long flags;
	rk_spin_lock(flags);
	{
		bzero(tmr, sizeof(struct rk_timer));
		list_add(&tmr->tmr_tmr, &rk_free_timer_root);	/* return to free list */
	}
	rk_spin_unlock(flags);
}

static inline void
rk_timer_add(rk_timer_t tmr)
{
	struct list_head	*lh;
	rk_timer_t		tmr1;

	/* find a place */
	lh = rk_online_timer_root.next;
	while (lh != &rk_online_timer_root) {
		tmr1 = list_entry(lh, struct rk_timer, tmr_tmr);
		if (tmr->tmr_expire < tmr1->tmr_expire) {
			break;
		}
		lh = lh->next;
	}
	/* the same code can be used to add a timer at the end 
	 * and to add a timer before `lh'. */
	list_add(&tmr->tmr_tmr, lh->prev);

	/* acutual (OS dependent) timer set will be done later at sowhere else. */
}

#ifdef	linux
/*
 *
 *
 * How to set the first linux timer??
 *	<-- when a reserve is created, "" is called and 
 *	    "" flag is set to invoke a bottom half.
 *	    "rk_timer_add()" is called too often to put such code in it.
 */
extern	unsigned long volatile jiffies;
extern	void	rk_replenish_timer_process(unsigned long);
static	struct timer_list	rk_linux_timer = {
	NULL, NULL, 0, 0,
	rk_replenish_timer_process
};

static inline void
rk_linux_timer_set(rk_timer_t tmr)
{
	struct timer_list	*tm = &rk_linux_timer;
#ifdef	DEBUG_RK_LINUX_TIMER2
	cpu_tick_data_t	tick = tmr->tmr_expire - rk_cpu_tick_at_last_jiffy;
	printk("rk_linux_timer_set: last intr(%lu usec) expire(%lu usec) TICK2USEC(%lu) TICK2JIFFIES(%lu)\n",
	       TICK2USEC(&rk_cpu_tick_at_last_jiffy),
	       TICK2USEC(&tmr->tmr_expire),
	       TICK2USEC(&tick),
	       TICK2JIFFIES(&tmr->tmr_expire)
		);
#endif

	/* delete a timer from queue if it is on it */
	del_timer(tm);

	/* expires = jiffies + TIMEOUT;
	 * convert it to jiffies */
	tm->expires = TICK2JIFFIES(&tmr->tmr_expire);

	/* enqueue "timer_list" */
	add_timer(tm);


#ifdef	DEBUG_RK_LINUX_TIMER
	printk("rk_linux_timer_set: rsv(0x%x) tmr(0x%x) expire(%lu) current(%lu)\n",
	       (int)tmr->tmr_rsv, (int)tmr, tm->expires, jiffies);
#endif
}

static inline void
rk_linux_timer_update(rk_timer_t tmr)
{
	struct list_head	*lh;
	rk_timer_t		tmr1;

	lh = rk_online_timer_root.next;
	if (lh != &rk_online_timer_root) {
		tmr1 = list_entry(lh, struct rk_timer, tmr_tmr);
		if (tmr1 == tmr) {
			/* timer at top of queue is the one just added */
			/* set a linux timer */
			rk_linux_timer_set(tmr);
		} else {
			/* there must be a timer already set */
		}
	} else {
		/* empty?? */
	}
}
#endif

/*
 * rk_replenish_reserve(rk_timer_t tmr)
 */
void
rk_replenish_reserve(rk_timer_t tmr)
{
	rk_reserve_t		rsv = tmr->tmr_rsv;
	cpu_tick_data_t		period;

#ifdef	DEBUG_RK
	printk("rk_replenish_reserve: rsv(0x%x) tmr(0x%x) expired(%lu usec)\n",
	       (int)rsv, (int)tmr, TICK2USEC(&tmr->tmr_expire));
#endif

	if (tmr->tmr_type != TMR_REPLENISH_RSV) {
		/* something wrong */
		rk_timer_destroy(tmr);
	}

	/* replenish it, and set timer expiration `ticks' later */
	(rsv->rsv_ops->replenish)(rsv, &period);
	tmr->tmr_expire += period;

#ifdef	DEBUG_RK
	printk("rk_replenish_reserve: next expire(%lu usec)\n",
	       TICK2USEC(&tmr->tmr_expire));
#endif
	/* set next timer */
	/*tmr->tmr_handler = rk_replenish_reserve;*/
	rk_timer_add(tmr);
}

/*
 * called from a bottom half.
 *
 *	lock is held by the callee.
 */
void
rk_replenish_timer_process(unsigned long bogus)
{
	struct list_head	*lh;
	rk_timer_t		tmr;
	cpu_tick_data_t		now;
	unsigned long		flags;

#ifdef	DEBUG_RK
	rk_rdtsc(&now);
	printk("rk_replenish_timer_process: current jiffies(%lu) (%lu usec)\n",
	       jiffies, TICK2USEC(&now));
#endif
	lh = rk_online_timer_root.next;
	while (lh != &rk_online_timer_root) {
		rk_spin_lock(flags);
		/* */
		tmr = list_entry(lh, struct rk_timer, tmr_tmr);
		rk_rdtsc(&now);
		if (tmr->tmr_expire < now) {	/* tmr expired */
			list_del(&tmr->tmr_tmr);
#ifdef	DEBUG_RK
			printk("rk_replenish_timer_process: expired tmr(0x%x) (%lu usec) < now(%lu usec) calling handler (0x%x)\n",
			       (int)tmr, TICK2USEC(&tmr->tmr_expire),
			       TICK2USEC(&now),
			       (int)tmr->tmr_handler);
#endif
			tmr->tmr_handler(tmr);
		} else {
#ifdef	linux
#ifdef	DEBUG_RK
			printk("rk_replenish_timer_process: set next timer for tmr(0x%x) (%lu usec)\n",
			       (int)tmr, TICK2USEC(&tmr->tmr_expire));
#endif
			rk_linux_timer_set(tmr);
#endif
			break;
		}
		lh = rk_online_timer_root.next;
		/* */
		rk_spin_unlock(flags);
	}
}

/*
 * rk_replenish_timer_create():
 * 	Create a timer to replenish a reserve, add it to the queue.
 * 	Set a linux timer if necessary.
 */
void
rk_replenish_timer_create(rk_reserve_t rsv, cpu_tick_t ticks)
{
	rk_timer_t		tmr;
	unsigned long flags;

	/* create a timer for a reserve and add it to queue*/
	tmr = rk_timer_create();
	rsv->rsv_replenish_tmr = tmr;
	tmr->tmr_rsv = rsv;
	tmr->tmr_type = TMR_REPLENISH_RSV;

	rk_spin_lock(flags);
	{
		/* the start time should be able to be specified. */
		rk_rdtsc(&tmr->tmr_expire);
		tmr->tmr_expire += *ticks;	/* obtain the timer expires. */

		tmr->tmr_handler = rk_replenish_reserve;
		rk_timer_add(tmr);
#ifdef	linux
		rk_linux_timer_update(tmr);
#endif
	}
	rk_spin_unlock(flags);

#ifdef	DEBUG_RK
	printk("rk_replenish_timer_create: rsv(0x%x) replenish(0x%x)\n",
	       (int)rsv, (int)rsv->rsv_replenish_tmr);
#endif
}

void
rk_replenish_timer_destroy(rk_reserve_t rsv)
{
	rk_timer_t	replenish_tmr;
	unsigned long flags;

#ifdef	DEBUG_RK
	printk("rk_replenish_timer_destory: rsv(0x%x) replenish(0x%x)\n",
	       (int)rsv, (int)rsv->rsv_replenish_tmr);
#endif
	replenish_tmr = rsv->rsv_replenish_tmr;

	rk_spin_lock(flags);
	{
		list_del(&replenish_tmr->tmr_tmr);
	}
	rk_spin_unlock(flags);

	rk_timer_destroy(replenish_tmr);
}


/*
 * Enforcement timer management
 *
 * cpu_reserve_start_account(rsv)
 *	 -> rk_timer_set_enforce_timer(rsv, ticks_expired);
 *
 * cpu_reserve_stop_account(rsv)
 *	 -> rk_timer_cancel_enforce_timer(rsv);
 *
 * "rk_current_resource_set != 0" --> accounting is on
 * "rk_current_resource_set == 0" --> accounting is off
 *
 */
#include <linux/timex.h>
#include <asm/io.h>

static	cpu_tick_data_t	rk_enforcement_expire;
static	cpu_tick_data_t	rk_cpu_tick_at_next_jiffy;
static	cpu_tick_data_t	rk_current_ISA_timer_setting;

void
get_rk_current_ISA_timer_setting(cpu_tick_t *t)
{
	*t = &rk_current_ISA_timer_setting;
}

static inline void
rk_program_ISA_timer(cpu_tick_t now, cpu_tick_t next)
{
	unsigned int		latch;
	cpu_tick_data_t		c;
	int			err=0;

	/* calculate latch value*/
	/*c = ((*next)-(*now))*CLOCK_TICK_RATE + rk_cpu_ticks_per_second/2;*/
	c = ((*next)-(*now))*CLOCK_TICK_RATE;
	c /= rk_cpu_ticks_per_second;
	latch = (unsigned int)c;
	latch = (LATCH>=latch)?latch:(err=1,LATCH);

	if (latch > LATCH) {
		err=1;
		latch = LATCH;
	} else if (latch == 0) {
		err=2;
		latch = 1;
	}

	/* set it */
	outb_p(latch & 0xff, 0x40);	/* LSB */
	outb(latch >> 8, 0x40);		/* MSB */

	rk_current_ISA_timer_setting = *next;

	switch (err) {
	case 1:
		printk("rk_program_ISA_timer: latch > LATCH\n");
	case 2:
		printk("rk_program_ISA_timer: latch == 0\n");
	}

#ifdef	UPERF_RK
	uperf_rk_isa_timer_programmed = 1;
#endif

#ifdef	DEBUG_RK_ENFORCE_TIMER_START
#ifdef	DEBUG_RK_ENFORCE
	DEBUG_RK_ENFORCE
#endif
	{
		printk("rk_program_ISA_timer: latch(%d) next(%lu usec)\n",
		       latch, TICK2USEC(next));
	}
#endif
}

static inline void
rk_enforce_timer_set_next_jiffy(cpu_tick_t now)
{
	if (rk_cpu_tick_at_next_jiffy != rk_current_ISA_timer_setting) {
		if (*now < rk_cpu_tick_at_next_jiffy) {
			rk_program_ISA_timer(now, &rk_cpu_tick_at_next_jiffy);
		} else {
			/* because now in Critical Section (I hope) */
			cpu_tick_data_t		t, ns, nt;
			t = *now - rk_cpu_tick_at_next_jiffy;
			ns = 9*1000*1000;
			nanosec2tick(&ns, &nt);
			if (t > nt) {
				printk("rk_enforce_timer_set_next_jiffy: (%lu usec)\n",
				       TICK2USEC(&t));
				/* this is too long. make it 9msec later */
				rk_rdtsc(&t);
				nt += t;
				rk_program_ISA_timer(&t, &nt);
			}
		}
	}
}

/*
 *
 */
static inline void
rk_enforce_timer_set(cpu_tick_t how_long, cpu_tick_t start_time)
{
	cpu_tick_data_t	now;

	rk_rdtsc(&now);
	rk_enforcement_expire = now + *how_long;

#if	HIGH_RESOLUTION_TIMER
	/*
	 * only this portion enables high resolution timer
	 */
	if (rk_enforcement_expire < rk_cpu_tick_at_next_jiffy) {
		rk_program_ISA_timer(&now, &rk_enforcement_expire);
	} else {
		rk_enforce_timer_set_next_jiffy(&now);
	}
#endif

	*start_time = now;
	return;
}

/*
 * Cancel enforce_timer and set timer to next jiffy
 */
void
rk_enforce_timer_cancel(void)
{
	cpu_tick_data_t	now;
	rk_rdtsc(&now);
	rk_enforce_timer_set_next_jiffy(&now);
	rk_enforcement_expire = 0;
}

/* 
 * rk_timer_isr():
 * 1) determines if this IRQ0 interrupts needs to be propagated to Linux.
 *    i.e. check if this is for a jiffy or not.
 * 2) set the timer to get a interrupt at next jiffy anyway
 *    since it may be a next event.
 */
int
rk_timer_isr(void)
{
	cpu_tick_data_t	now;
	int		no_IRQ = 0;
	int		ret = 0;

	rk_current_ISA_timer_setting = 0;	/* clear since alarmed */
	rk_rdtsc(&now);
	/*
	 * timer managemenet first
	 */
	if (now >= rk_cpu_tick_at_next_jiffy) {
		/* 
		 * this is a time for Linux jiffy.
		 * increment internal tick to count jiffies
		 */
		rk_cpu_tick_at_last_jiffy += rk_cpu_ticks_per_jiffy;
		rk_cpu_tick_at_next_jiffy
			= rk_cpu_tick_at_last_jiffy + rk_cpu_ticks_per_jiffy;
		ret = 1;	/* go to Linux */
		/*
		 */
		if (now >= rk_cpu_tick_at_next_jiffy) {
			no_IRQ = 1;
			/* this must not happen.
			 * So, I don't know what should be here */
			rk_cpu_tick_at_last_jiffy = now;
			rk_cpu_tick_at_next_jiffy
				= rk_cpu_tick_at_last_jiffy + rk_cpu_ticks_per_jiffy;
		}
	}
	/*
	 * Set the next timer alarm interrupt:
	 * if enforce timer is expired,
	 *	set it to rk_cpu_tick_at_next_jiffy
	 * if not expired,
	 *	set it to either rk_cpu_tick_at_next_jiffy or rk_enforcement_expire
	 * 	which comes sooner.
	 */
	if (now < rk_enforcement_expire) {
#if	HIGH_RESOLUTION_TIMER
		if (rk_enforcement_expire < rk_cpu_tick_at_next_jiffy) {
			rk_program_ISA_timer(&now, &rk_enforcement_expire);
		} else {
			rk_enforce_timer_set_next_jiffy(&now);
		}
#else
		rk_enforce_timer_set_next_jiffy(&now);
#endif
	} else {	/* (now >= rk_enforcement_expire) --> expired or not set */
		if (rk_enforcement_expire) {
			/*
			 * cancel and set jiffy
			 */
			rk_enforce_timer_cancel();
			/*
			 * When enforce timer expired, do enforcement.
			 */
			rk_resource_set_do_enforcement(rk_current_resource_set);
		} else {
			rk_enforce_timer_set_next_jiffy(&now);
		}
	}


	if (no_IRQ) {
		printk("rk_timer_isr: long no IRQ0\n");
	}

#ifdef	DEBUG_RK_ENFORCE_TIMER_ISR
#ifdef	DEBUG_RK_ENFORCE
	DEBUG_RK_ENFORCE
#endif
	{
		if (ret) {
			printk("rk_timer_isr: jiffies(%lu) now(%lu usec) last_jiffy(%lu usec) next_jiffy(%lu usec) alarm@(%lu usec)\n",
			       jiffies, TICK2USEC(&now),
			       TICK2USEC(&rk_cpu_tick_at_last_jiffy),
			       TICK2USEC(&rk_cpu_tick_at_next_jiffy),
			       TICK2USEC(&rk_current_ISA_timer_setting));
		}
	}
#endif

	return ret;
}

/* called from "rsv->rsv_ops->start_account()" */
void
rk_enforce_timer_start(rk_reserve_t rsv, cpu_tick_t next_available_ticks,
		       cpu_tick_t start)
{
#ifdef	DEBUG_RK_ENFORCE_TIMER_START
#ifdef	DEBUG_RK_ENFORCE
	DEBUG_RK_ENFORCE
#endif
	{
		printk("rk_enforce_timer_start: rsv(0x%x) next_available(%lu usec)\n",
		       (int)rsv, TICK2USEC(next_available_ticks));
	}
#endif
	rk_enforce_timer_set(next_available_ticks, start);
}

/*
 *
 */
void
rk_enforce_timer_set_oneshot_mode(void)
{
	cpu_tick_data_t	now;
	cpu_tick_data_t	ns = RK_NANOSECS_PER_JIFFY;

	ns *= (jiffies+1);
	nanosec2tick(&ns, &rk_cpu_tick_at_next_jiffy);
	rk_cpu_tick_at_next_jiffy += rk_cpu_tick_at_boot;

	rk_rdtsc(&now);
	while (now > rk_cpu_tick_at_next_jiffy) {
		rk_cpu_tick_at_next_jiffy += rk_cpu_ticks_per_jiffy;
	}
	
	rk_cpu_tick_at_last_jiffy
		= rk_cpu_tick_at_next_jiffy - rk_cpu_ticks_per_jiffy;

	/* */
	outb_p(0x38, 0x43);	/* binary, mode 4, LSB/MSB, ch 0 */
	/* */
	rk_enforce_timer_cancel();

#ifdef	DEBUG_RK_ENFORCE_TIMER
	printk("rk_enforce_timer_set_oneshot_mode: jiffies(%lu) now(%lu usec) last_jiffy(%lu usec) next_jiffy(%lu usec)\n",
	       jiffies,
	       TICK2USEC(&now),
	       TICK2USEC(&rk_cpu_tick_at_last_jiffy),
	       TICK2USEC(&rk_cpu_tick_at_next_jiffy));
#endif
}

void
rk_enforce_timer_set_periodic_mode(void)
{
	/* set the clock to 100 Hz */
	outb_p(0x34,0x43);		/* binary, mode 2, LSB/MSB, ch 0 */
	outb_p(LATCH & 0xff , 0x40);	/* LSB */
	outb(LATCH >> 8 , 0x40);	/* MSB */
}
